/*
 * This file is part of the CAF spoki driver.
 *
 * Copyright (C) 2018-2021
 * Authors: Raphael Hiesgen
 *
 * All rights reserved.
 *
 * Report any bugs, questions or comments to raphael.hiesgen@haw-hamburg.de
 *
 */

#define SUITE regression
#include "test.hpp"

#include <cmath>
#include <vector>

#include "spoki/analysis/regression.hpp"

namespace {

double round(double x, unsigned n) {
  unsigned mult = 1;
  for (unsigned i = 0; i < n; ++i)
    mult *= 10;
  x *= mult;
  x = static_cast<int>(x + 0.5) * 1.0;
  return x / mult;
}

struct regression_test {
  regression_test(std::vector<double> xs, std::vector<double> ys)
    : xs_{std::move(xs)}, ys_{std::move(ys)} {
    std::tie(m_, b_) = spoki::analysis::least_squares_method(xs_, ys_);
    std::vector<double> predictions(xs_.size());
    std::transform(std::begin(xs_), std::end(xs_), std::begin(predictions),
                   [&](const int x) { return m_ * x + b_; });
    sigma_ = spoki::analysis::sigma_for_prediction(ys_, predictions);
    std_err_ = spoki::analysis::std_err_for_prediciton(xs_, 0.0, sigma_);
    interval_ = spoki::analysis::prediction_interval(std_err_, 3.182);
  }

  static regression_test make(std::vector<double> xs, std::vector<double> ys) {
    return regression_test(std::move(xs), std::move(ys));
  }

  regression_test& with_lobf(double m, double b) {
    CHECK_EQUAL(m, m_);
    CHECK_EQUAL(b, b_);
    return *this;
  }

  regression_test& check_interval(double sigma, double std_err,
                                  double interval) {
    CHECK_EQUAL(sigma, sigma_);
    CHECK_EQUAL(std_err, std_err_);
    CHECK_EQUAL(interval, interval_);
    return *this;
  }

  std::vector<double> xs_;
  std::vector<double> ys_;
  double m_;
  double b_;
  double sigma_;
  double std_err_;
  double interval_;
};

} // namespace

TEST(correlation coefficient) {
  // ys = 50 - xs**2
  std::vector<double> xs = {-7, -6, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 6, 7};
  std::vector<double> ys = {1,  14, 25, 34, 41, 46, 49, 50,
                            49, 46, 41, 34, 25, 14, 1};
  CHECK_EQUAL(xs.size(), ys.size());
  CHECK_EQUAL(spoki::analysis::correlation_coefficient(xs, ys), 0.0);
}

TEST(prediction interval m) {
  // Computer repair example ... from a book.
  std::vector<double> xs = {1, 2, 3, 4, 4, 5, 6, 6, 7, 8, 9, 9, 10, 10};
  std::vector<double> ys = {23, 29,  49,  64,  74,  87,  96,
                            97, 109, 119, 149, 145, 154, 166};
  MESSAGE("covariance and correlation coefficient");
  CHECK_EQUAL(xs.size(), ys.size());
  CHECK_EQUAL(spoki::analysis::covariance(xs, ys), 136.0);
  auto cc = round(spoki::analysis::correlation_coefficient(xs, ys, 3), 3);
  CHECK_EQUAL(cc, 0.994);
  MESSAGE("line of best fit with least squares method");
  double m, b;
  std::tie(m, b) = spoki::analysis::least_squares_method(xs, ys);
  CHECK_EQUAL(round(m, 3), 15.509);
  CHECK_EQUAL(round(b, 3), 4.162);
  MESSAGE("line of best fit again");
  std::tie(m, b) = spoki::analysis::line_of_best_fit(xs, ys);
  CHECK_EQUAL(round(m, 3), 15.509);
  CHECK_EQUAL(round(b, 3), 4.162);
  std::vector<double> predictions(xs.size());
  std::transform(std::begin(xs), std::end(xs), std::begin(predictions),
                 [m, b](const int x) { return m * x + b; });
  auto sigma = spoki::analysis::sigma_for_prediction(ys, predictions);
  CHECK_EQUAL(sigma, 5.3917249166919206);
  auto se_m = spoki::analysis::std_err_m(xs, sigma);
  auto se_b = spoki::analysis::std_err_b(xs, sigma);
  CHECK_EQUAL(round(se_m, 3), 0.505);
  CHECK_EQUAL(round(se_b, 3), 3.355);
  auto interval = spoki::analysis::prediction_interval(se_m, 2.18);
  auto lower = m - interval;
  auto upper = m + interval;
  CHECK_EQUAL(round(lower, 3), 14.408);
  CHECK_EQUAL(round(upper, 3), 16.610);
}

TEST(prediction interval by hand comparison) {
  MESSAGE("Let's start with a bit of often used values for the data points.");
  // X:   133,  1119,  2100,  3116
  // Y: 34212, 34215, 34221, 34227
  std::vector<double> xs = {133, 1119, 2100, 3116};
  std::vector<double> ys = {34212, 34215, 34221, 34227};
  //  xbar = (133+1119+2100+3116) / 4
  //       = 1617
  //  ybar = (34212+34215+34221+34227) / 4
  //       = 34218.75
  auto x_bar = spoki::analysis::mean(xs);
  auto y_bar = spoki::analysis::mean(ys);
  CHECK_EQUAL(x_bar, 1617);
  CHECK_EQUAL(y_bar, 34218.75);
  MESSAGE("Now for the line of best fit.");
  //  SXX = sum((xi - xbar)^2)
  //      = (133 - 1617)^2 + (1119 - 1617)^2 + (2100 - 1617)^2 + (3116 - 1617)^2
  //      = 4930550.00
  //  SYY = sum((yi - ybar)^2)
  //      = (34212 - 34218.75)^2 + (34215 - 34218.75)^2 + (34221 - 34218.75)^2 +
  //      (34227 - 34218.75)^2 = 132.75
  //  SXY = sum((xi - xbar)(yi - ybar))
  //      = (133 - 1617)(34212 - 34218.75) + (1119 - 1617)(34215 - 34218.75) +
  //      (2100 - 1617)(34221 - 34218.75) + (3116 - 1617)(34227 - 34218.75) =
  //      25338.00
  //  m = SXY / SXX = 25338.00 / 4930550.00
  //    = 0.005138980438
  //  b = ybar - m * xbar = 34218.75 - 0.005138980438 * 1617
  //    = 34210.4402686318
  double m, b;
  std::tie(m, b) = spoki::analysis::least_squares_method(xs, ys);
  CHECK_EQUAL(m, 0.005138980438);
  CHECK_EQUAL(b, 34210.4402686318);
  MESSAGE("Next up: variance (sigma squared).");
  //  RSS = SYY - (SXY^2 / SXX) = 132.75 - (25338.00 ^ 2 / 4930550.00)
  //      = 2.5385136547
  //  Var = RSS / (n - 2) = 2.5385136547 / 2
  //      = 1.2692568274
  //  sigma = sqrt(Var) = sqrt(1.2692568274)
  //        = 1.1266129892
  std::vector<double> predictions(xs.size());
  std::transform(std::begin(xs), std::end(xs), std::begin(predictions),
                 [m, b](const int x) { return m * x + b; });
  auto sigma = spoki::analysis::sigma_for_prediction(ys, predictions);
  CHECK_EQUAL(sigma, 1.1266129892);

  MESSAGE(
    "So far so good, now for the standard error of the prediction at x = 0.");
  //  sepred = sgima * sqrt(1 + (1 / n) + (x - xbar)^2 / SXX)
  //         = 1.1266129892 * sqrt(1 + (1/4) + (0 - 1617)^2 /  4930550.00)
  //         = 1.5032174326
  auto std_err = spoki::analysis::std_err_for_prediciton(xs, 0.0, sigma);
  CHECK_EQUAL(std_err, 1.5032174326);

  MESSAGE("Assuming t = 3.18 we can now calculate the prediction interval.");
  // for n = 3 (sample size - 1), which doesn't play into this, right?
  //  interval = 3.18 * 1.5032174326 = 4.7802314357
  //  34210.4402686318 - 4.7802314357, 34210.4402686318 + 4.7802314357
  auto interval = spoki::analysis::prediction_interval(std_err, 3.18);
  CHECK_EQUAL(interval, 4.7802314357);
}

TEST(python generated events) {
  using dvec = std::vector<double>;
  regression_test::make(dvec{1, 1002, 2002, 3003},
                        dvec{19265, 19385, 19433, 19661})
    .with_lobf(0.1235309134914448, 19250.45656793585)
    .check_interval(55.15875823722129, 71.93217747616637, 228.88818872916139);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{9957, 10544, 11277, 12133})
    .with_lobf(0.7258685200077174, 9887.67695007841)
    .check_interval(95.43625574220758, 124.46219967742326, 396.0387193735608);

  regression_test::make(dvec{1, 1002, 2002, 3003},
                        dvec{39981, 40138, 40272, 40367})
    .with_lobf(0.12912220170906843, 39995.55845303298)
    .check_interval(22.07052928773862, 28.782033541995833, 91.58443073063074);

  regression_test::make(dvec{1, 1002, 2002, 3003},
                        dvec{21596, 21746, 21824, 22019})
    .with_lobf(0.13462299855253754, 21594.04625617409)
    .check_interval(33.81726590425619, 44.10087628010042, 140.32898832327953);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{47039, 47169, 47188, 47204})
    .with_lobf(0.05139135747681029, 47072.8230289092)
    .check_interval(43.75338094239144, 57.060516394566626, 181.566563167511);

  regression_test::make(dvec{1, 1002, 2002, 3003},
                        dvec{11157, 11158, 11160, 11162})
    .with_lobf(0.0016989605677344916, 11156.698161227263)
    .check_interval(0.38751798114251423, 0.5053596760620852,
                    1.6080544892295552);

  regression_test::make(dvec{1, 1002, 2002, 3003}, dvec{6567, 6568, 6570, 6572})
    .with_lobf(0.0016989605677344916, 6566.698161227263)
    .check_interval(0.3875179811432177, 0.5053596760630027, 1.6080544892324746);

  regression_test::make(dvec{1, 1001, 2002, 3002},
                        dvec{19998, 20215, 20589, 21217})
    .with_lobf(0.40293686991669264, 19899.740289820085)
    .check_interval(146.13101906511497, 190.56179264450742, 606.3676241948226);

  regression_test::make(dvec{1, 1002, 2003, 3003},
                        dvec{34618, 35043, 35607, 36027})
    .with_lobf(0.47876791160140664, 34604.52090479679)
    .check_interval(44.699120575369115, 58.292859158950826, 185.48787784378152);

  regression_test::make(dvec{1, 1002, 2002, 3003},
                        dvec{58727, 58732, 58898, 59082})
    .with_lobf(0.12302332280330691, 58674.968969149435)
    .check_interval(67.21689346797154, 87.65711311956234, 278.92493394644737);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{43882, 44033, 44196, 44444})
    .with_lobf(0.18484041802312343, 43861.16590223378)
    .check_interval(36.23758932583079, 47.258874978110974, 150.37774018034912);

  regression_test::make(dvec{1, 1001, 2002, 3002}, dvec{130, 202, 245, 448})
    .with_lobf(0.09965635498474121, 106.61598299041108)
    .check_interval(55.136508177575806, 71.90062661368003, 228.78779388472987);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{11195, 11196, 11198, 11200})
    .with_lobf(0.001699430087033267, 11194.697880866797)
    .check_interval(0.3879565326949864, 0.505949474473036, 1.6099312277732003);

  regression_test::make(dvec{1, 1002, 2002, 3002}, dvec{6607, 6608, 6610, 6612})
    .with_lobf(0.001699430087033267, 6606.697880866798)
    .check_interval(0.38795653269498626, 0.5059494744730357,
                    1.6099312277731996);

  regression_test::make(dvec{1, 1002, 2002, 3003},
                        dvec{52724, 53133, 53806, 54429})
    .with_lobf(0.5784466326621523, 52654.17315774145)
    .check_interval(90.59578789061709, 118.1453771151699, 375.93858998047057);

  regression_test::make(dvec{1, 1001, 2002, 3002},
                        dvec{16762, 17578, 17689, 18690})
    .with_lobf(0.5892323962365316, 16795.017557050847)
    .check_interval(260.7158795172044, 339.98589546241186, 1081.8351193613946);

  regression_test::make(dvec{1, 1002, 2002, 3003},
                        dvec{60687, 60749, 60827, 60905})
    .with_lobf(0.07315578379712992, 60682.12001273671)
    .check_interval(6.206253676034815, 8.093535009738389, 25.75362840098755);

  regression_test::make(dvec{1, 1002, 2003, 3003},
                        dvec{11242, 11268, 11270, 11272})
    .with_lobf(0.009194522611356665, 11249.18752840709)
    .check_interval(9.292786518073871, 12.118875913428505, 38.562263156529504);

  regression_test::make(dvec{2, 1002, 2003, 3004},
                        dvec{10328, 10353, 10355, 10357})
    .with_lobf(0.00889239575512516, 10334.886952278986)
    .check_interval(8.911305973309018, 11.62240606269119, 36.98249609148337);

  regression_test::make(dvec{1, 1002, 2003, 3003},
                        dvec{57368, 57531, 57643, 57815})
    .with_lobf(0.14519679462928195, 57371.12811526816)
    .check_interval(17.868567470786786, 23.30269306284183, 74.1491693259627);

  regression_test::make(dvec{1, 1002, 2002, 3003}, dvec{2677, 2812, 2906, 3038})
    .with_lobf(0.11763099575269846, 2681.5682443794467)
    .check_interval(12.49888415160006, 16.29971344134144, 51.865688170348456);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{11294, 11295, 11297, 11299})
    .with_lobf(0.001699430087033267, 11293.697880866797)
    .check_interval(0.3879565326949864, 0.505949474473036, 1.6099312277732003);

  regression_test::make(dvec{1, 1002, 2003, 3003}, dvec{6687, 6688, 6690, 6692})
    .with_lobf(0.0016987707865748626, 6686.6980215858675)
    .check_interval(0.38773727064273067, 0.5056545591348259, 1.608992807167016);

  regression_test::make(dvec{1, 1002, 2002, 3003},
                        dvec{63686, 64082, 64702, 64893})
    .with_lobf(0.4238326313093762, 63704.153387773316)
    .check_interval(126.25784989947667, 164.6520399837386, 523.9227912282562);

  regression_test::make(dvec{1, 1002, 2003, 3003},
                        dvec{35983, 36735, 36851, 36894})
    .with_lobf(0.2847304707392062, 36188.013650332025)
    .check_interval(265.92651440756777, 346.7991461902055, 1103.5148831772337);

  regression_test::make(dvec{1, 1002, 2002, 3003},
                        dvec{62318, 62448, 62621, 62667})
    .with_lobf(0.12192344309890407, 62330.370988465445)
    .check_interval(40.0820934335221, 52.27079707046066, 166.3256762782058);

  regression_test::make(dvec{2, 1003, 2003, 3003},
                        dvec{12980, 13126, 13221, 13392})
    .with_lobf(0.13306009398459515, 12979.79394376465)
    .check_interval(21.939447343523813, 28.61715289067108, 91.05978049811537);

  regression_test::make(dvec{2, 1002, 2002, 3003},
                        dvec{17048, 17147, 17255, 17337})
    .with_lobf(0.09746955364737853, 17050.326363033226)
    .check_interval(8.205866560088339, 10.702539947368074, 34.05548211252521);

  regression_test::make(dvec{1, 1002, 2003, 3003},
                        dvec{11334, 11335, 11337, 11339})
    .with_lobf(0.0016987707865748626, 11333.698021585868)
    .check_interval(0.3877372706424954, 0.5056545591345192, 1.60899280716604);

  regression_test::make(dvec{1, 1002, 2002, 3003}, dvec{6725, 6726, 6728, 6730})
    .with_lobf(0.0016989605677344916, 6724.698161227263)
    .check_interval(0.3875179811432177, 0.5053596760630027, 1.6080544892324746);

  regression_test::make(dvec{1, 1002, 2002, 3003},
                        dvec{56570, 56572, 57060, 57622})
    .with_lobf(0.36417324643497945, 56409.01178385466)
    .check_interval(208.4673052785678, 271.86085547439103, 865.0612421195123);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{33889, 34339, 34676, 34722})
    .with_lobf(0.28353333746751885, 33980.70381045815)
    .check_interval(145.492266064077, 189.7422245828886, 603.7597586227515);

  regression_test::make(dvec{1, 1002, 2002, 3003},
                        dvec{17437, 17546, 17704, 17873})
    .with_lobf(0.14651132780210635, 17419.939985641235)
    .check_interval(22.060346576315734, 28.768754334329312, 91.54217629183587);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{24276, 24470, 24532, 24771})
    .with_lobf(0.15465443414135008, 24279.997703528228)
    .check_interval(51.376103791547926, 67.00161106512988, 213.19912640924326);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{29359, 29430, 29435, 29545})
    .with_lobf(0.05628287183446854, 29357.727197222586)
    .check_interval(30.351573513715923, 39.58268871520085, 125.9521154917691);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{11372, 11373, 11375, 11377})
    .with_lobf(0.001699430087033267, 11371.697880866797)
    .check_interval(0.3879565326949864, 0.505949474473036, 1.6099312277732003);

  regression_test::make(dvec{1, 1003, 2003, 3003}, dvec{6764, 6765, 6767, 6769})
    .with_lobf(0.001698860348265872, 6763.697462326731)
    .check_interval(0.3886143127660118, 0.5068340923384902, 1.6127460818210757);

  regression_test::make(dvec{1, 1001, 2002, 3002},
                        dvec{36708, 37324, 37446, 37555})
    .with_lobf(0.26618389964350525, 36858.57487468528)
    .check_interval(194.75102249448668, 253.96458741838566, 808.1153171653032);

  regression_test::make(dvec{1, 1002, 2003, 3003},
                        dvec{52175, 52789, 52999, 53263})
    .with_lobf(0.34716987123683923, 52284.96406093446)
    .check_interval(143.29756215911573, 186.87671035222277, 594.6416923407728);

  regression_test::make(dvec{1, 1002, 2002, 3003},
                        dvec{18999, 19048, 19095, 19303})
    .with_lobf(0.09584574676558259, 18967.289688358094)
    .check_interval(61.82870417082461, 80.63041053988006, 256.5659663378984);

  regression_test::make(dvec{1, 1001, 2002, 3003},
                        dvec{43717, 43727, 43815, 43996})
    .with_lobf(0.09244367800676762, 43674.922706553334)
    .check_interval(60.47204993482202, 78.85564069465386, 250.91864869038858);

  regression_test::make(dvec{2, 1002, 2003, 3003},
                        dvec{52614, 52805, 52907, 53005})
    .with_lobf(0.1274473166572109, 52641.26040672254)
    .check_interval(35.53638886741326, 46.34931856910778, 147.48353168690096);

  regression_test::make(dvec{2, 1002, 2003, 3003},
                        dvec{11410, 11411, 11413, 11415})
    .with_lobf(0.001699340187981577, 11409.696741367557)
    .check_interval(0.3870792628771681, 0.5048588401462757, 1.6064608293454492);

  regression_test::make(dvec{2, 1002, 2003, 3003}, dvec{6803, 6804, 6806, 6808})
    .with_lobf(0.001699340187981577, 6802.6967413675575)
    .check_interval(0.3870792628769339, 0.5048588401459702, 1.606460829344477);

  regression_test::make(dvec{1, 1002, 2002, 3003},
                        dvec{45035, 45552, 45805, 46620})
    .with_lobf(0.5005161803770753, 45001.22469707363)
    .check_interval(167.6785111305343, 218.66845460349413, 695.8030225483183);

  regression_test::make(dvec{2, 1002, 2003, 3003},
                        dvec{14127, 14443, 14699, 14726})
    .with_lobf(0.20522128193019945, 14190.405023899875)
    .check_interval(105.59680866537856, 137.7275598534674, 438.2490954537333);

  regression_test::make(dvec{1, 1002, 2002, 3003},
                        dvec{31391, 31481, 31624, 31730})
    .with_lobf(0.11592863926143059, 31382.375183829332)
    .check_interval(15.347465909078219, 20.01453036403312, 63.686235618353386);

  regression_test::make(dvec{2, 1003, 2003, 3003},
                        dvec{60546, 60585, 60631, 60743})
    .with_lobf(0.06367783374807609, 60530.55813533508)
    .check_interval(27.461334682417228, 35.819736061883866, 113.97840014891446);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{60460, 60560, 60807, 60814})
    .with_lobf(0.13086151346269537, 60463.7287221574)
    .check_interval(69.46108414283131, 90.5869499715915, 288.2476748096041);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{11467, 11468, 11470, 11472})
    .with_lobf(0.001699430087033267, 11466.697880866797)
    .check_interval(0.3879565326949864, 0.505949474473036, 1.6099312277732003);

  regression_test::make(dvec{1, 1002, 2002, 3002}, dvec{6845, 6846, 6848, 6850})
    .with_lobf(0.001699430087033267, 6844.697880866798)
    .check_interval(0.38795653269498626, 0.5059494744730357,
                    1.6099312277731996);

  regression_test::make(dvec{2, 1002, 2002, 3003}, dvec{65027, 118, 292, 340})
    .with_lobf(-19.38028976909508, 45558.290305623086)
    .check_interval(25179.027567483794, 32839.86479716485, 104496.44978457857);

  regression_test::make(dvec{2, 1002, 2003, 3004}, dvec{6416, 6878, 7402, 7609})
    .with_lobf(0.4100040188713697, 6460.116460641049)
    .check_interval(108.33994460523364, 141.30036975309127, 449.61777655433644);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{34832, 34983, 35191, 35388})
    .with_lobf(0.18754074742043564, 34816.86068256136)
    .check_interval(19.56782193351827, 25.519171323322475, 81.20200315081212);

  regression_test::make(dvec{1, 1002, 2002, 3002}, dvec{3974, 4023, 4111, 4129})
    .with_lobf(0.05528387138458854, 3976.2274461481943)
    .check_interval(20.421148500093402, 26.632028284161446, 84.74311400020171);

  regression_test::make(dvec{1, 1001, 2002, 3002},
                        dvec{64468, 64610, 64754, 64909})
    .with_lobf(0.146641157745572, 64465.06830164502)
    .check_interval(4.825362938804201, 6.292502561479853, 20.02274315062889);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{11505, 11506, 11508, 11510})
    .with_lobf(0.001699430087033267, 11504.697880866797)
    .check_interval(0.3879565326949864, 0.505949474473036, 1.6099312277732003);

  regression_test::make(dvec{1, 1002, 2003, 3003}, dvec{6883, 6884, 6886, 6888})
    .with_lobf(0.0016987707865748626, 6882.6980215858675)
    .check_interval(0.38773727064273067, 0.5056545591348259, 1.608992807167016);

  regression_test::make(dvec{1, 1002, 2002, 3003},
                        dvec{59413, 59700, 60385, 60528})
    .with_lobf(0.4027395514424485, 59401.585193733445)
    .check_interval(157.2255533359764, 205.03682040325884, 652.4271625231696);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{21552, 22519, 22995, 23261})
    .with_lobf(0.5601697641160658, 21740.5150567387)
    .check_interval(251.58131986918684, 328.0971599855026, 1044.0051630738694);

  regression_test::make(dvec{1, 1002, 2002, 3003},
                        dvec{41260, 41408, 41463, 41522})
    .with_lobf(0.08405150457390947, 41287.00464012999)
    .check_interval(34.993362459705224, 45.63461614044739, 145.2093485589036);

  regression_test::make(dvec{1, 1002, 2003, 3004},
                        dvec{17849, 17893, 17933, 18032})
    .with_lobf(0.05884115884115884, 17838.341158841158)
    .check_interval(21.848340898110813, 28.49177458768714, 90.66082673802048);

  regression_test::make(dvec{1, 1002, 2002, 3003},
                        dvec{15379, 15597, 15761, 15892})
    .with_lobf(0.1701982939679209, 15401.612162460182)
    .check_interval(30.93211083574641, 40.33836434062277, 128.35667533186165);

  regression_test::make(dvec{2, 1003, 2003, 3003},
                        dvec{11544, 11545, 11547, 11549})
    .with_lobf(0.001699430087033267, 11543.696181436711)
    .check_interval(0.38795653269545594, 0.5060387910982497,
                    1.6102154332746303);

  regression_test::make(dvec{2, 1002, 2003, 3003}, dvec{6923, 6924, 6926, 6928})
    .with_lobf(0.001699340187981577, 6922.6967413675575)
    .check_interval(0.3870792628769339, 0.5048588401459702, 1.606460829344477);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{31760, 32698, 33074, 34033})
    .with_lobf(0.7192945653665952, 31811.049386460716)
    .check_interval(181.08954845131797, 236.16604993100492, 751.4803708804576);

  regression_test::make(dvec{1, 1001, 2002, 3002},
                        dvec{60904, 61647, 62490, 63046})
    .with_lobf(0.7266170610277656, 60930.73448286681)
    .check_interval(89.92932150028366, 117.2721084546884, 373.15984910281844);

  regression_test::make(dvec{1, 1002, 2002, 3003},
                        dvec{46960, 47001, 47128, 47152})
    .with_lobf(0.07025406701795166, 46954.728391339035)
    .check_interval(30.503742904129986, 39.7797325101234, 126.57910884721267);

  regression_test::make(dvec{1, 1002, 2002, 3003},
                        dvec{27888, 27911, 27918, 28126})
    .with_lobf(0.0720610978581311, 27852.51423101709)
    .check_interval(73.84975856715293, 96.30698930854126, 306.4488399797783);

  regression_test::make(dvec{1, 1002, 2002, 3003},
                        dvec{19396, 19642, 19870, 19882})
    .with_lobf(0.1684949386756138, 19444.42060210923)
    .check_interval(88.47564115693582, 115.38050756414067, 367.1407750690956);

  regression_test::make(dvec{3973, 4974, 5975, 6976},
                        dvec{11599, 11601, 11603, 11605})
    .with_lobf(0.001998001998001998, 11591.061938061937)
    .check_interval(0.0, 0.0, 0.0);

  regression_test::make(dvec{1, 1002, 2002, 3003},
                        dvec{11627, 11628, 11630, 11632})
    .with_lobf(0.0016989605677344916, 11626.698161227263)
    .check_interval(0.38751798114251423, 0.5053596760620852,
                    1.6080544892295552);

  regression_test::make(dvec{2, 1002, 2003, 3003}, dvec{7002, 7003, 7005, 7007})
    .with_lobf(0.001699340187981577, 7001.6967413675575)
    .check_interval(0.3870792628769339, 0.5048588401459702, 1.606460829344477);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{61735, 61909, 62559, 63410})
    .with_lobf(0.567293188992626, 61551.317453430325)
    .check_interval(243.48637801720997, 317.5402258170104, 1010.4129985497271);

  regression_test::make(dvec{2, 1002, 2003, 3002},
                        dvec{18091, 18823, 19362, 20145})
    .with_lobf(0.6700172460482035, 18098.716592124085)
    .check_interval(71.77722545338358, 93.62074668780107, 297.901215960583);

  regression_test::make(dvec{1, 1002, 2002, 3002}, dvec{559, 746, 765, 847})
    .with_lobf(0.0882810681169696, 596.6739059553408)
    .check_interval(52.046270637342545, 67.87560218233996, 215.98016614420575);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{43523, 43698, 43700, 43828})
    .with_lobf(0.09167782955059521, 43549.57281947239)
    .check_interval(50.086959717816626, 65.32038723808537, 207.84947219158767);

  regression_test::make(dvec{2, 1002, 2002, 3003},
                        dvec{39071, 39269, 39321, 39397})
    .with_lobf(0.10296470573185502, 39109.82127081432)
    .check_interval(50.84551958917152, 66.31550739498448, 211.0159445308406);

  regression_test::make(dvec{1, 1002, 2002, 3003},
                        dvec{11665, 11666, 11668, 11670})
    .with_lobf(0.0016989605677344916, 11664.698161227263)
    .check_interval(0.38751798114251423, 0.5053596760620852,
                    1.6080544892295552);

  regression_test::make(dvec{1, 1002, 2002, 3002}, dvec{7040, 7041, 7043, 7045})
    .with_lobf(0.001699430087033267, 7039.697880866798)
    .check_interval(0.38795653269498626, 0.5059494744730357,
                    1.6099312277731996);

  regression_test::make(dvec{1, 1002, 2002, 3003},
                        dvec{37381, 38290, 38739, 38798})
    .with_lobf(0.469719548653796, 37596.481237921995)
    .check_interval(300.7186958464146, 392.1652932612376, 1247.869963157258);

  regression_test::make(dvec{1, 1002, 2002, 3003},
                        dvec{59813, 60186, 60996, 61126})
    .with_lobf(0.47459289868441923, 59817.411466176)
    .check_interval(196.53596566261507, 256.30127316668097, 815.5506512163788);

  regression_test::make(dvec{1, 1002, 2002, 3003},
                        dvec{14526, 14770, 14848, 14952})
    .with_lobf(0.13552251876847032, 14570.445176809757)
    .check_interval(58.04308607397495, 75.693610628097, 240.85706901860468);

  regression_test::make(dvec{1, 1001, 2002, 3003},
                        dvec{64898, 65140, 65212, 65293})
    .with_lobf(0.12560223778586901, 64947.12683940507)
    .check_interval(63.61864694083993, 82.95880774761235, 263.9749262529025);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{24885, 25132, 25802, 25868})
    .with_lobf(0.36179022159850766, 24878.43153471444)
    .check_interval(174.54391011441413, 227.62962381738467, 724.317462986918);

  regression_test::make(dvec{2, 1002, 2002, 3003},
                        dvec{48489, 48568, 48577, 48614})
    .with_lobf(0.03838736182480085, 48504.33258569869)
    .check_interval(21.4666064717215, 27.99792217137346, 89.08938834931035);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{11722, 11723, 11725, 11727})
    .with_lobf(0.001699430087033267, 11721.697880866797)
    .check_interval(0.3879565326949864, 0.505949474473036, 1.6099312277732003);

  regression_test::make(dvec{1, 1002, 2002, 3002}, dvec{7079, 7080, 7082, 7084})
    .with_lobf(0.001699430087033267, 7078.697880866798)
    .check_interval(0.38795653269498626, 0.5059494744730357,
                    1.6099312277731996);

  regression_test::make(dvec{1, 1002, 2002, 3003},
                        dvec{23375, 23711, 24566, 25545})
    .with_lobf(0.7360504450457671, 23193.702231541258)
    .check_interval(235.8195885133524, 307.5307899490075, 978.5629736177418);

  regression_test::make(dvec{1, 1001, 2002, 3002},
                        dvec{33222, 34215, 34472, 35430})
    .with_lobf(0.6877961255403426, 33302.024117501176)
    .check_interval(227.76355212558053, 297.0144947308637, 945.1001222336083);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{28737, 28812, 28988, 29207})
    .with_lobf(0.1585446494286454, 28697.90557272053)
    .check_interval(51.79064885598629, 67.54223569655703, 214.91939398644445);

  regression_test::make(dvec{1, 1002, 2002, 3003}, dvec{5465, 5633, 5706, 5803})
    .with_lobf(0.10863719191482545, 5488.576937743932)
    .check_interval(31.350572749354555, 40.88407779753604, 130.09313555175967);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{53965, 54046, 54279, 54498})
    .with_lobf(0.1831364906351453, 53921.97477518867)
    .check_interval(55.47290740703921, 72.34441486296328, 230.19992809394915);

  regression_test::make(dvec{1, 1002, 2002, 3003},
                        dvec{11761, 11762, 11764, 11766})
    .with_lobf(0.0016989605677344916, 11760.698161227263)
    .check_interval(0.38751798114251423, 0.5053596760620852,
                    1.6080544892295552);

  regression_test::make(dvec{1, 1002, 2002, 3003}, dvec{7117, 7118, 7120, 7122})
    .with_lobf(0.0016989605677344916, 7116.698161227263)
    .check_interval(0.3875179811432177, 0.5053596760630027, 1.6080544892324746);

  regression_test::make(dvec{1, 1002, 2003, 3003},
                        dvec{56096, 56626, 57450, 57803})
    .with_lobf(0.5941005824292787, 56101.26240004561)
    .check_interval(136.00658190695998, 177.36842295196004, 564.3863218331369);

  regression_test::make(dvec{2, 1002, 2002, 3003},
                        dvec{54503, 55168, 55295, 56089})
    .with_lobf(0.4883719535720637, 54530.09323274637)
    .check_interval(195.79482174657645, 255.36631456117496, 812.5756129336587);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{39975, 40218, 40349, 40524})
    .with_lobf(0.17775162236383835, 39999.56150111511)
    .check_interval(34.38017685112322, 44.836549830162504, 142.66990155957708);

  regression_test::make(dvec{1, 1002, 2002, 3003},
                        dvec{21185, 21295, 21359, 21542})
    .with_lobf(0.11343523234709055, 21174.87028101467)
    .check_interval(36.67260056962304, 47.824499626001284, 152.17755780993608);

  regression_test::make(dvec{1, 1002, 2002, 3002}, dvec{7901, 8117, 8178, 8269})
    .with_lobf(0.1164731486398438, 7941.336449030115)
    .check_interval(52.95314000154418, 69.05828642545914, 219.743467405811);

  regression_test::make(dvec{1, 1002, 2002, 3003},
                        dvec{11800, 11801, 11803, 11805})
    .with_lobf(0.0016989605677344916, 11799.698161227263)
    .check_interval(0.38751798114251423, 0.5053596760620852,
                    1.6080544892295552);

  regression_test::make(dvec{1, 1001, 2002, 3002}, dvec{7156, 7157, 7159, 7161})
    .with_lobf(0.001699340187981577, 7155.698440707745)
    .check_interval(0.3870792628769339, 0.5047697518380748, 1.606177350348754);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{9616, 9673, 10419, 10949})
    .with_lobf(0.47432498386093586, 9451.93245548684)
    .check_interval(220.2710570155867, 287.26420654521036, 914.0747052268594);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{57716, 57891, 58424, 59012})
    .with_lobf(0.4419437174779556, 57597.06102227748)
    .check_interval(153.84637414460286, 200.6371476910815, 638.4274039530213);

  regression_test::make(dvec{2, 1003, 2003, 3003},
                        dvec{62066, 62301, 62399, 62591})
    .with_lobf(0.16725427234845008, 62087.908642228365)
    .check_interval(39.514863979282545, 51.541995850820534, 164.00663079731095);

  regression_test::make(dvec{1, 1002, 2002, 3003},
                        dvec{45427, 45496, 45551, 45758})
    .with_lobf(0.10474046954036374, 45400.67981475037)
    .check_interval(55.38648240063512, 72.22915107495274, 229.83315872049963);

  regression_test::make(dvec{2, 1003, 2003, 3003},
                        dvec{13444, 13550, 13721, 13722})
    .with_lobf(0.10047275127832031, 13458.264573016504)
    .check_interval(52.50994839311495, 68.49239171426022, 217.94279043477601);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{11838, 11839, 11841, 11843})
    .with_lobf(0.001699430087033267, 11837.697880866797)
    .check_interval(0.3879565326949864, 0.505949474473036, 1.6099312277732003);

  regression_test::make(dvec{2, 1002, 2003, 3003}, dvec{7195, 7196, 7198, 7200})
    .with_lobf(0.001699340187981577, 7194.6967413675575)
    .check_interval(0.3870792628769339, 0.5048588401459702, 1.606460829344477);

  regression_test::make(dvec{1, 1001, 2002, 3002},
                        dvec{46512, 47073, 47803, 48781})
    .with_lobf(0.7533970316952374, 46411.0243569096)
    .check_interval(147.98026423728004, 192.97329622055483, 614.0410285738054);

  regression_test::make(dvec{1, 1002, 2002, 3003},
                        dvec{63699, 64675, 65087, 65364})
    .with_lobf(0.54038432265908, 63894.59274736606)
    .check_interval(256.2284076697492, 334.145798128185, 1063.2519296438845);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{11649, 11785, 11888, 12021})
    .with_lobf(0.12186421319245255, 11652.740417838235)
    .check_interval(9.99384604110188, 13.033370303974962, 41.47218430724833);

  regression_test::make(dvec{1, 1002, 2002, 3003},
                        dvec{64313, 64358, 64401, 64592})
    .with_lobf(0.08795022455044964, 64283.89876272523)
    .check_interval(56.79508378879526, 74.06609897378952, 235.67832693459826);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{17344, 17546, 17668, 17917})
    .with_lobf(0.1840444956959081, 17342.36117858867)
    .check_interval(36.709845319232734, 47.874762716921936, 152.33749496524558);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{11895, 11896, 11898, 11900})
    .with_lobf(0.001699430087033267, 11894.697880866797)
    .check_interval(0.3879565326949864, 0.505949474473036, 1.6099312277732003);

  regression_test::make(dvec{1, 1002, 2003, 3003}, dvec{7234, 7235, 7237, 7239})
    .with_lobf(0.0016987707865748626, 7233.6980215858675)
    .check_interval(0.38773727064273067, 0.5056545591348259, 1.608992807167016);

  regression_test::make(dvec{2, 1002, 2003, 3003},
                        dvec{18920, 19342, 19468, 20086})
    .with_lobf(0.3622393360832662, 18909.735397534892)
    .check_interval(142.66799992642407, 186.07868691663754, 592.1023817687407);

  regression_test::make(dvec{1, 1002, 2002, 3002}, dvec{8458, 8592, 8818, 9569})
    .with_lobf(0.3557667365930337, 8324.977303321411)
    .check_interval(228.73684752312928, 298.304779128267, 949.2058071861455);

  regression_test::make(dvec{1, 1002, 2003, 3003},
                        dvec{26471, 26563, 26790, 26962})
    .with_lobf(0.16987897600016297, 26441.299308303755)
    .check_interval(41.2830239831652, 53.837871343543, 171.31210661515382);

  regression_test::make(dvec{1, 1001, 2002, 3002},
                        dvec{19404, 19609, 19756, 19845})
    .with_lobf(0.14694121763764637, 19432.867761717072)
    .check_interval(41.01221963235141, 53.48188319954626, 170.1793523409562);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{24784, 24842, 24873, 25003})
    .with_lobf(0.06877702346936486, 24772.21410500488)
    .check_interval(32.337368700089975, 42.17244284043154, 134.19271311825318);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{11934, 11935, 11937, 11939})
    .with_lobf(0.001699430087033267, 11933.697880866797)
    .check_interval(0.3879565326949864, 0.505949474473036, 1.6099312277732003);

  regression_test::make(dvec{2, 1002, 2003, 3003}, dvec{7300, 7301, 7306, 7308})
    .with_lobf(0.002898980236015141, 7299.394282195387)
    .check_interval(1.161021792699835, 1.5142948017677509, 4.818486059224983);

  regression_test::make(dvec{1, 1001, 2002, 3002}, dvec{4614, 4833, 5656, 6576})
    .with_lobf(0.6706418523897177, 4412.781258636839)
    .check_interval(260.4177292152351, 339.597093301281, 1080.597950884676);

  regression_test::make(dvec{1, 1001, 2002, 3002},
                        dvec{10119, 10583, 11348, 11607})
    .with_lobf(0.5227070298347263, 10129.405394703159)
    .check_interval(146.60216749420746, 191.17619258384886, 608.3226448018071);

  regression_test::make(dvec{1, 1002, 2003, 3003},
                        dvec{47584, 47657, 47727, 47886})
    .with_lobf(0.09752650924747346, 47566.990801482985)
    .check_interval(33.74384294627198, 44.005901213025055, 140.02677765984572);

  regression_test::make(dvec{2, 1002, 2003, 3004},
                        dvec{28951, 29015, 29075, 29248})
    .with_lobf(0.09503774488639935, 28929.432028871965)
    .check_interval(42.723735706227714, 55.721642414719284, 177.30626616363676);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{44762, 44996, 45178, 45268})
    .with_lobf(0.16995580102601424, 44795.76887580918)
    .check_interval(51.24671648696935, 66.83287195845661, 212.66219857180891);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{11997, 11998, 12000, 12002})
    .with_lobf(0.001699430087033267, 11996.697880866797)
    .check_interval(0.3879565326949864, 0.505949474473036, 1.6099312277732003);

  regression_test::make(dvec{1, 1002, 2002, 3002}, dvec{7344, 7345, 7347, 7349})
    .with_lobf(0.001699430087033267, 7343.697880866798)
    .check_interval(0.38795653269498626, 0.5059494744730357,
                    1.6099312277731996);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{14819, 15631, 15654, 16172})
    .with_lobf(0.40810507573879534, 14956.128202509264)
    .check_interval(227.95773907713067, 297.288712956956, 945.9726846290339);

  regression_test::make(dvec{2, 1002, 2002, 3003},
                        dvec{41462, 42146, 43005, 43851})
    .with_lobf(0.8023654603689596, 41410.64648716073)
    .check_interval(64.33694970936197, 83.91176840546193, 267.00724706617984);

  regression_test::make(dvec{1, 1002, 2002, 3002}, dvec{1235, 1324, 1485, 1596})
    .with_lobf(0.12436036512686911, 1223.2418216707242)
    .check_interval(20.83374458889007, 27.170111179289357, 86.45529377249873);

  regression_test::make(dvec{2, 1002, 2002, 3003},
                        dvec{30606, 30768, 30936, 31028})
    .with_lobf(0.1433526668969618, 30619.14845615404)
    .check_interval(27.994634252242882, 36.51213302125188, 116.18160727362347);

  regression_test::make(dvec{1, 1002, 2003, 3003},
                        dvec{62644, 62655, 62850, 62979})
    .with_lobf(0.11991265632917816, 62601.86121202949)
    .check_interval(57.489483394760214, 74.97298191808835, 238.56402846335712);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{12036, 12037, 12039, 12041})
    .with_lobf(0.001699430087033267, 12035.697880866797)
    .check_interval(0.3879565326949864, 0.505949474473036, 1.6099312277732003);

  regression_test::make(dvec{1, 1002, 2002, 3002}, dvec{7383, 7384, 7386, 7388})
    .with_lobf(0.001699430087033267, 7382.697880866798)
    .check_interval(0.38795653269498626, 0.5059494744730357,
                    1.6099312277731996);

  regression_test::make(dvec{2, 1002, 2003, 3003},
                        dvec{26244, 26272, 26696, 27413})
    .with_lobf(0.392944865518612, 26065.850339558285)
    .check_interval(244.13381191242615, 318.4182800350852, 1013.2069670716411);

  regression_test::make(dvec{1, 1002, 2002, 3003}, dvec{1264, 1655, 2277, 2383})
    .with_lobf(0.3976464651831941, 1297.4850092948425)
    .check_interval(155.34822355517193, 202.58860685948486, 644.6369470268808);

  regression_test::make(dvec{2, 1003, 2003, 3003},
                        dvec{10069, 10105, 10343, 10489})
    .with_lobf(0.1497466295602694, 10026.468252428305)
    .check_interval(60.66071916596163, 79.12401108603163, 251.77260327575263);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{50296, 50459, 50703, 50918})
    .with_lobf(0.21093300855491576, 50277.231354402655)
    .check_interval(25.385211620298136, 33.10585953910044, 105.3428450534176);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{64465, 64659, 64700, 64886})
    .with_lobf(0.13036426188822778, 64481.72546970935)
    .check_interval(47.17943854702197, 61.528573763096574, 195.7839217141733);

  regression_test::make(dvec{2, 1002, 2002, 3003},
                        dvec{12074, 12075, 12077, 12079})
    .with_lobf(0.0016995300270542634, 12073.696881016858)
    .check_interval(0.3868598340896983, 0.5045637530245932, 1.6055218621242555);

  regression_test::make(dvec{1, 1002, 2002, 3002}, dvec{7423, 7424, 7426, 7428})
    .with_lobf(0.001699430087033267, 7422.697880866798)
    .check_interval(0.38795653269498626, 0.5059494744730357,
                    1.6099312277731996);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{42944, 43229, 44188, 44830})
    .with_lobf(0.6614737665190236, 42804.38177113006)
    .check_interval(201.4324455538942, 262.6960274694925, 835.8987594079251);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{40487, 41150, 41861, 42296})
    .with_lobf(0.6136240335362732, 40526.9901076369)
    .check_interval(95.38040042097496, 124.38935654154963, 395.8069325152109);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{34384, 34591, 34728, 34774})
    .with_lobf(0.13066862922217232, 34423.0183860656)
    .check_interval(56.97394583617072, 74.30197850832145, 236.42889561347886);

  regression_test::make(dvec{2, 1003, 2003, 3004},
                        dvec{54240, 54244, 54383, 54460})
    .with_lobf(0.0798481502804034, 54211.738230128554)
    .check_interval(40.471269870842264, 52.78763324995464, 167.97024900135565);

  regression_test::make(dvec{1, 1002, 2002, 3002}, dvec{1132, 1259, 1269, 1324})
    .with_lobf(0.058587638628677, 1158.0160136893844)
    .check_interval(36.091154671151614, 47.067903747223575, 149.77006972366542);

  regression_test::make(dvec{1, 1002, 2002, 3003},
                        dvec{12130, 12131, 12133, 12135})
    .with_lobf(0.0016989605677344916, 12129.698161227263)
    .check_interval(0.38751798114251423, 0.5053596760620852,
                    1.6080544892295552);

  regression_test::make(dvec{1, 1002, 2002, 3002}, dvec{7461, 7462, 7464, 7466})
    .with_lobf(0.001699430087033267, 7460.697880866798)
    .check_interval(0.38795653269498626, 0.5059494744730357,
                    1.6099312277731996);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{9400, 9582, 9987, 10151})
    .with_lobf(0.26571653022438585, 9380.960200735528)
    .check_interval(73.67414421052385, 96.08136841162907, 305.7309142858037);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{26703, 27566, 27844, 28151})
    .with_lobf(0.462095273521596, 26872.048422988944)
    .check_interval(219.0628159575948, 285.6884915441477, 909.060780093478);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{52457, 52674, 52745, 52914})
    .with_lobf(0.14416158142690663, 52481.00534509214)
    .check_interval(42.10599084968105, 54.912089750300765, 174.73026958545702);

  regression_test::make(dvec{1, 1002, 2002, 3003},
                        dvec{11519, 11601, 11816, 11896})
    .with_lobf(0.13451392947905336, 11505.960077922462)
    .check_interval(42.42295197875493, 55.32349545787039, 176.03936254694358);

  regression_test::make(dvec{2, 1002, 2002, 3003}, dvec{8870, 9031, 9183, 9395})
    .with_lobf(0.1726514332423396, 8860.384384411695)
    .check_interval(21.008511230448224, 27.40044930442817, 87.18822968669043);

  regression_test::make(dvec{1, 1002, 2002, 3003},
                        dvec{12169, 12170, 12172, 12174})
    .with_lobf(0.0016989605677344916, 12168.698161227263)
    .check_interval(0.38751798114251423, 0.5053596760620852,
                    1.6080544892295552);

  regression_test::make(dvec{1, 1001, 2001, 3002},
                        dvec{43771, 43862, 44002, 44114})
    .with_lobf(0.11686521334221267, 43761.806098470006)
    .check_interval(14.254418175053141, 18.588110863380017, 59.14736876727521);

  regression_test::make(dvec{2, 1003, 2004, 3004}, dvec{7499, 7500, 7502, 7504})
    .with_lobf(0.0016987707865748626, 7498.696322815082)
    .check_interval(0.3877372706427306, 0.5057437852042281, 1.6092767245198538);

  regression_test::make(dvec{2, 1002, 2003, 3003},
                        dvec{42287, 42978, 43514, 43834})
    .with_lobf(0.5174942011402476, 42375.71496278678)
    .check_interval(131.51053215994327, 171.52625082464658, 545.7965301240254);

  regression_test::make(dvec{2, 1002, 2002, 3003},
                        dvec{47103, 47508, 48460, 49332})
    .with_lobf(0.7636879226931956, 46953.49981813414)
    .check_interval(192.41678528460585, 250.96049466230534, 798.5562940154556);

  regression_test::make(dvec{1, 1002, 2002, 3002}, dvec{3855, 4087, 4329, 4465})
    .with_lobf(0.2071414840383544, 3872.9252763454015)
    .check_interval(38.530980963480665, 50.24977781397603, 159.89479300407172);

  regression_test::make(dvec{1, 1001, 2002, 3002},
                        dvec{33727, 33905, 34139, 34189})
    .with_lobf(0.16194001559951726, 33746.84706657733)
    .check_interval(59.0264018849297, 76.97323284617092, 244.92882691651585);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{24681, 24776, 24870, 24964})
    .with_lobf(0.09427177279555675, 24681.177365204272)
    .check_interval(0.35078704453249687, 0.457475273325794, 1.4556863197226766);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{12208, 12209, 12211, 12213})
    .with_lobf(0.001699430087033267, 12207.697880866797)
    .check_interval(0.3879565326949864, 0.505949474473036, 1.6099312277732003);

  regression_test::make(dvec{1, 1001, 2002, 3002}, dvec{7537, 7538, 7540, 7542})
    .with_lobf(0.001699340187981577, 7536.698440707745)
    .check_interval(0.3870792628769339, 0.5047697518380748, 1.606177350348754);

  regression_test::make(dvec{1, 1002, 2002, 3003}, dvec{7035, 7204, 7635, 7887})
    .with_lobf(0.29851206611583453, 6991.884876694016)
    .check_interval(75.73852050202203, 98.77011144999568, 314.28649463388626);

  regression_test::make(dvec{1, 1001, 2002, 3002},
                        dvec{38705, 39409, 39732, 39831})
    .with_lobf(0.36994886691669326, 38863.77177632458)
    .check_interval(215.34895381284963, 280.825268620141, 893.5860047492887);

  regression_test::make(dvec{1, 1001, 2002, 3002},
                        dvec{25907, 26014, 26083, 26127})
    .with_lobf(0.07287058895471843, 25923.33481068449)
    .check_interval(22.370633941635262, 29.172369656934514, 92.82648024836563);

  regression_test::make(dvec{1, 1001, 2002, 3003},
                        dvec{46401, 46501, 46585, 46636})
    .with_lobf(0.07884252709867196, 46412.34823492957)
    .check_interval(17.557042753898333, 22.89440917836082, 72.85001000554414);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{34217, 34307, 34350, 34521})
    .with_lobf(0.09546905424709361, 34205.37934778443)
    .check_interval(39.835276760151935, 51.950761602801784, 165.30732342011527);

  regression_test::make(dvec{1, 1002, 2002, 3003},
                        dvec{12246, 12247, 12249, 12251})
    .with_lobf(0.0016989605677344916, 12245.698161227263)
    .check_interval(0.38751798114251423, 0.5053596760620852,
                    1.6080544892295552);

  regression_test::make(dvec{1, 1002, 2002, 3003}, dvec{7577, 7578, 7580, 7582})
    .with_lobf(0.0016989605677344916, 7576.698161227263)
    .check_interval(0.3875179811432177, 0.5053596760630027, 1.6080544892324746);

  regression_test::make(dvec{1, 1001, 2002, 3002}, dvec{2271, 2809, 3308, 3472})
    .with_lobf(0.4100418844840359, 2349.32211044722)
    .check_interval(140.22403381986808, 182.8588031994169, 581.8567117805446);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{63308, 63318, 64229, 64872})
    .with_lobf(0.5600886128190167, 63090.636925699044)
    .check_interval(290.46785184877365, 378.8106259567514, 1205.375411794383);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{26828, 26904, 27097, 27099})
    .with_lobf(0.1005704426488443, 26830.9683377521)
    .check_interval(55.2791981646958, 72.09179097058183, 229.39607886839136);

  regression_test::make(dvec{1, 1002, 2002, 3002}, dvec{4998, 5204, 5300, 5513})
    .with_lobf(0.16405269377583043, 5007.3838671221465)
    .check_interval(35.95523800851531, 46.89064944619333, 149.20604653778716);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{39115, 39319, 39490, 39661})
    .with_lobf(0.1808477142442963, 39124.66194513363)
    .check_interval(12.710803405035545, 16.576661973530722, 52.746938399774756);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{12336, 12337, 12339, 12341})
    .with_lobf(0.001699430087033267, 12335.697880866797)
    .check_interval(0.3879565326949864, 0.505949474473036, 1.6099312277732003);

  regression_test::make(dvec{2, 1002, 2003, 3003}, dvec{7646, 7647, 7649, 7651})
    .with_lobf(0.001699340187981577, 7645.6967413675575)
    .check_interval(0.3870792628769339, 0.5048588401459702, 1.606460829344477);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{32717, 32745, 33009, 33492})
    .with_lobf(0.2587994315211726, 32602.09795371308)
    .check_interval(160.9814268036262, 209.9422523585894, 668.0362470050314);

  regression_test::make(dvec{1, 1002, 2002, 3002}, dvec{1858, 2267, 2538, 3085})
    .with_lobf(0.39507869352208275, 1843.6905720032123)
    .check_interval(81.67542787404075, 106.51615922838486, 338.9344186647206);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{46872, 46920, 46927, 47147})
    .with_lobf(0.0831689861329723, 46841.600975074805)
    .check_interval(72.89321744680436, 95.0629309001864, 302.49024612439314);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{21139, 21276, 21524, 21606})
    .with_lobf(0.1648505149634447, 21138.685739153647)
    .check_interval(47.92046078041791, 62.49497019662249, 198.85899516565274);

  regression_test::make(dvec{2, 1002, 2003, 3003},
                        dvec{63338, 63373, 63454, 63700})
    .with_lobf(0.11665095590508476, 63290.98193875261)
    .check_interval(76.94504221407064, 100.35770058684092, 319.3382032673278);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{12376, 12377, 12379, 12381})
    .with_lobf(0.001699430087033267, 12375.697880866797)
    .check_interval(0.3879565326949864, 0.505949474473036, 1.6099312277732003);

  regression_test::make(dvec{2, 1002, 2003, 3003}, dvec{7685, 7686, 7688, 7690})
    .with_lobf(0.001699340187981577, 7684.6967413675575)
    .check_interval(0.3870792628769339, 0.5048588401459702, 1.606460829344477);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{47559, 48025, 48088, 48424})
    .with_lobf(0.26573352002795525, 47624.93468629802)
    .check_interval(116.2724941592692, 151.63556315676814, 482.5043619648362);

  regression_test::make(dvec{1, 1002, 2003, 3003},
                        dvec{20765, 20927, 21389, 21651})
    .with_lobf(0.311781734092317, 20714.625889959818)
    .check_interval(86.60262456287424, 112.93990869298968, 359.37478946109314);

  regression_test::make(dvec{1, 1002, 2003, 3003}, dvec{4064, 4081, 4181, 4270})
    .with_lobf(0.07174711457867683, 4041.217897124183)
    .check_interval(29.49330477400998, 38.46270439313176, 122.38832537894525);

  regression_test::make(dvec{1, 1002, 2003, 3004},
                        dvec{27794, 27917, 28051, 28091})
    .with_lobf(0.1023976023976024, 27809.397602397603)
    .check_interval(33.715723334966434, 43.96767671286885, 139.9051473003487);

  regression_test::make(dvec{1, 1002, 2002, 3003},
                        dvec{11451, 11453, 11632, 11862})
    .with_lobf(0.14111280818505467, 11387.548562106047)
    .check_interval(83.04625234526705, 108.30007696582224, 344.6108449052464);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{12415, 12416, 12418, 12420})
    .with_lobf(0.001699430087033267, 12414.697880866797)
    .check_interval(0.3879565326949864, 0.505949474473036, 1.6099312277732003);

  regression_test::make(dvec{1, 1002, 2002, 3003}, dvec{7726, 7727, 7729, 7731})
    .with_lobf(0.0016989605677344916, 7725.698161227263)
    .check_interval(0.3875179811432177, 0.5053596760630027, 1.6080544892324746);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{63656, 64255, 65152, 252})
    .with_lobf(-18.923208686306914, 76746.67864466141)
    .check_interval(25411.139248128628, 33139.67278509489, 105450.43880217194);

  regression_test::make(dvec{1, 1002, 2002, 3002}, dvec{5113, 5746, 6448, 7069})
    .with_lobf(0.6568020202675364, 5107.647566063227)
    .check_interval(24.156165250928396, 31.50301150773476, 100.24258261761202);

  regression_test::make(dvec{2, 1002, 2003, 3003},
                        dvec{10810, 10924, 11069, 11281})
    .with_lobf(0.1557369792691888, 10787.005188648043)
    .check_interval(35.12069304332637, 45.807135787105445, 145.75830607456953);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{43096, 43206, 43347, 43590})
    .with_lobf(0.16224537843613154, 43066.09800293354)
    .check_interval(48.39393272631732, 63.11244370722804, 200.82379587639963);

  regression_test::make(dvec{1, 1001, 2002, 3002},
                        dvec{22537, 22726, 22775, 22870})
    .with_lobf(0.10475437554868594, 22569.711305113648)
    .check_interval(44.3998824703322, 57.89955651360932, 184.23638882630485);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{12453, 12454, 12456, 12458})
    .with_lobf(0.001699430087033267, 12452.697880866797)
    .check_interval(0.3879565326949864, 0.505949474473036, 1.6099312277732003);

  regression_test::make(dvec{1, 1002, 2002, 3003}, dvec{7765, 7766, 7768, 7770})
    .with_lobf(0.0016989605677344916, 7764.698161227263)
    .check_interval(0.3875179811432177, 0.5053596760630027, 1.6080544892324746);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{8667, 9578, 10334, 10838})
    .with_lobf(0.7267013202026804, 8762.926292385624)
    .check_interval(144.46803380150885, 188.40648273731412, 599.5094280701335);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{22006, 22821, 23484, 23809})
    .with_lobf(0.6070404846531354, 22118.376952172155)
    .check_interval(175.52418513811114, 228.90803928736847, 728.3853810124065);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{25006, 25162, 25317, 25350})
    .with_lobf(0.11866933060123967, 25030.53833276959)
    .check_interval(47.478637729955445, 61.918771263616485, 197.02553016082766);

  regression_test::make(dvec{2, 1002, 2002, 3003},
                        dvec{63580, 63815, 64019, 64044})
    .with_lobf(0.15954015197779053, 63624.83080669137)
    .check_interval(77.90802675091587, 101.61190928668005, 323.3290953502159);

  regression_test::make(dvec{1, 1002, 2003, 3003},
                        dvec{26840, 27085, 27125, 27368})
    .with_lobf(0.16228241536316312, 26860.711241520687)
    .check_interval(64.53939424988769, 84.16688674823301, 267.81903363287745);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{12492, 12493, 12495, 12497})
    .with_lobf(0.001699430087033267, 12491.697880866797)
    .check_interval(0.3879565326949864, 0.505949474473036, 1.6099312277732003);

  regression_test::make(dvec{1, 1001, 2002, 3002}, dvec{7804, 7805, 7807, 7809})
    .with_lobf(0.001699340187981577, 7803.698440707745)
    .check_interval(0.3870792628769339, 0.5047697518380748, 1.606177350348754);

  regression_test::make(dvec{1, 1002, 2002, 3003}, dvec{7571, 7679, 7917, 8575})
    .with_lobf(0.3248108969992421, 7447.634032707138)
    .check_interval(199.76383836375942, 260.51072093942844, 828.9451140292613);

  regression_test::make(dvec{1, 1002, 2003, 3003},
                        dvec{40587, 41581, 42500, 43469})
    .with_lobf(0.9558308610875039, 40598.3530889313)
    .check_interval(21.653799710525483, 28.23907675439471, 89.85674223248397);

  regression_test::make(dvec{2, 1002, 2002, 3003},
                        dvec{48393, 48513, 48760, 48876})
    .with_lobf(0.169546346759992, 48380.799000579806)
    .check_interval(40.84679945314593, 53.27462976252936, 169.51987190436842);

  regression_test::make(dvec{2, 1003, 2004, 3004}, dvec{3351, 3372, 3479, 3503})
    .with_lobf(0.05626215204295624, 3341.673919941426)
    .check_interval(26.73419649322776, 34.870657923769286, 110.95843351343386);

  regression_test::make(dvec{1, 1001, 2002, 3002},
                        dvec{31705, 31856, 31929, 32152})
    .with_lobf(0.1413389006117305, 31698.279640731485)
    .check_interval(44.16813983714772, 57.597353108101586, 183.27477758997924);

  regression_test::make(dvec{1, 1001, 2002, 3002},
                        dvec{12549, 12550, 12552, 12554})
    .with_lobf(0.001699340187981577, 12548.698440707745)
    .check_interval(0.38707926287763833, 0.5047697518389934, 1.606177350351677);

  regression_test::make(dvec{1, 1002, 2002, 3002}, dvec{7843, 7844, 7846, 7848})
    .with_lobf(0.001699430087033267, 7842.697880866798)
    .check_interval(0.38795653269498626, 0.5059494744730357,
                    1.6099312277731996);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{21237, 22021, 22412, 23164})
    .with_lobf(0.6170239930505707, 21281.884218436306)
    .check_interval(119.63594492603421, 156.02197246927378, 496.4619163972292);

  regression_test::make(dvec{1, 1002, 2002, 3003},
                        dvec{49784, 50640, 50815, 51306})
    .with_lobf(0.47383560773647315, 49924.54891717982)
    .check_interval(203.6077897155652, 265.5235928691895, 844.8960725097609);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{49315, 49377, 49593, 49729})
    .with_lobf(0.1457502280006654, 49284.6195951)
    .check_interval(45.36303489371319, 59.159729842777786, 188.24626035971892);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{18690, 18805, 18913, 19002})
    .with_lobf(0.10436986242706439, 18695.762559100156)
    .check_interval(9.353387365762064, 12.198122788076992, 38.814426711660985);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{39822, 39973, 40140, 40253})
    .with_lobf(0.14595740366419094, 39827.808469047304)
    .check_interval(17.3818139058614, 22.668311704839436, 72.13056784479909);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{12588, 12589, 12591, 12593})
    .with_lobf(0.001699430087033267, 12587.697880866797)
    .check_interval(0.3879565326949864, 0.505949474473036, 1.6099312277732003);

  regression_test::make(dvec{1, 1002, 2002, 3003}, dvec{7883, 7884, 7886, 7888})
    .with_lobf(0.0016989605677344916, 7882.698161227263)
    .check_interval(0.3875179811432177, 0.5053596760630027, 1.6080544892324746);

  regression_test::make(dvec{1, 1002, 2002, 3003},
                        dvec{19346, 19615, 20551, 20638})
    .with_lobf(0.4808811502672192, 19315.216512298637)
    .check_interval(248.33410358922322, 323.8508875769169, 1030.4935242697495);

  regression_test::make(dvec{2, 1002, 2002, 3003}, dvec{64815, 85, 350, 552})
    .with_lobf(-19.244030395158347, 45359.844661126626)
    .check_interval(25161.161021344287, 32816.56227056643, 104422.30114494238);

  regression_test::make(dvec{1, 1002, 2002, 3003},
                        dvec{12614, 12643, 12645, 12647})
    .with_lobf(0.010094482583107238, 12622.088087160173)
    .check_interval(10.455752283500088, 13.635278475069757, 43.387456107671966);

  regression_test::make(dvec{1, 1001, 2002, 3002},
                        dvec{54500, 54524, 54658, 54745})
    .with_lobf(0.08686838791598962, 54476.317115544145)
    .check_interval(33.3314707970725, 43.4658217482074, 138.30824480279594);

  regression_test::make(dvec{1, 1002, 2002, 3003},
                        dvec{19265, 19385, 19433, 19661})
    .with_lobf(0.12353091349144486, 19250.45656793585)
    .check_interval(55.15875823722129, 71.93217747616637, 228.88818872916139);

  regression_test::make(dvec{2, 1002, 2002, 3003},
                        dvec{47242, 47410, 47545, 47762})
    .with_lobf(0.16945275293032885, 47235.18960191042)
    .check_interval(25.054025420690262, 32.67683016093624, 103.97767357209912);

  regression_test::make(dvec{1, 1001, 2002, 3002},
                        dvec{12669, 12670, 12672, 12674})
    .with_lobf(0.001699340187981577, 12668.698440707745)
    .check_interval(0.38707926287763833, 0.5047697518389934, 1.606177350351677);

  regression_test::make(dvec{1, 1002, 2003, 3003}, dvec{7963, 7964, 7966, 7968})
    .with_lobf(0.0016987707865748626, 7962.6980215858675)
    .check_interval(0.38773727064273067, 0.5056545591348259, 1.608992807167016);

  regression_test::make(dvec{1, 1002, 2003, 3003},
                        dvec{31784, 32244, 32879, 33412})
    .with_lobf(0.5515130284280351, 31751.239553043983)
    .check_interval(50.86081285625955, 66.32842352100818, 211.057043643848);

  regression_test::make(dvec{2, 1002, 2003, 3003},
                        dvec{47359, 47912, 48507, 48946})
    .with_lobf(0.5353897810971661, 47376.57685390151)
    .check_interval(50.931601047937264, 66.42895008305841, 211.37691916429185);

  regression_test::make(dvec{2, 1002, 2002, 3003},
                        dvec{12024, 12044, 12137, 12242})
    .with_lobf(0.07468123005977963, 11999.560122142697)
    .check_interval(31.540318185277133, 41.13660792054048, 130.8966864031598);

  regression_test::make(dvec{1, 1002, 2003, 3004},
                        dvec{38763, 38963, 39186, 39209})
    .with_lobf(0.15594405594405594, 38795.944055944055)
    .check_interval(71.82861546765216, 93.66957108523341, 298.0565751932127);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{50244, 50269, 50514, 50749})
    .with_lobf(0.17593441296206083, 50179.79049533422)
    .check_interval(82.74215887561431, 107.90732536212944, 343.36110930229586);

  regression_test::make(dvec{2, 1002, 2003, 3003},
                        dvec{12709, 12710, 12712, 12714})
    .with_lobf(0.001699340187981577, 12708.696741367557)
    .check_interval(0.3870792628771681, 0.5048588401462757, 1.6064608293454492);

  regression_test::make(dvec{1, 1002, 2003, 3003}, dvec{8002, 8003, 8005, 8007})
    .with_lobf(0.0016987707865748626, 8001.6980215858675)
    .check_interval(0.38773727064273067, 0.5056545591348259, 1.608992807167016);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{41816, 42471, 42814, 43024})
    .with_lobf(0.39660502749275023, 41935.64839996276)
    .check_interval(159.70755495939454, 208.2809456507018, 662.7499690605331);

  regression_test::make(dvec{2, 1002, 2002, 3003}, dvec{3638, 4050, 4095, 5044})
    .with_lobf(0.42621165907280745, 3566.473535157875)
    .check_interval(276.3111950091345, 360.3801720190824, 1146.7297073647203);

  regression_test::make(dvec{1, 1002, 2003, 3003},
                        dvec{26400, 26613, 26629, 26867})
    .with_lobf(0.1415954385085968, 26414.53825250046)
    .check_interval(66.86557019445158, 87.20049110039832, 277.47196268146746);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{53552, 53737, 53756, 53826})
    .with_lobf(0.08408268777493187, 53591.478823633995)
    .check_interval(53.16974273561379, 69.34076662681227, 220.64231940651663);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{53739, 53943, 53975, 54100})
    .with_lobf(0.11147314938939396, 53771.84519790448)
    .check_interval(50.319711432237526, 65.62392796409382, 208.81533878174653);

  regression_test::make(dvec{2, 1002, 2002, 3003},
                        dvec{12765, 12766, 12768, 12770})
    .with_lobf(0.0016995300270542634, 12764.696881016858)
    .check_interval(0.3868598340896983, 0.5045637530245932, 1.6055218621242555);

  regression_test::make(dvec{1, 1002, 2002, 3003}, dvec{8041, 8042, 8044, 8046})
    .with_lobf(0.0016989605677344916, 8040.698161227263)
    .check_interval(0.3875179811432177, 0.5053596760630027, 1.6080544892324746);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{46606, 47409, 48347, 48424})
    .with_lobf(0.6390345833947756, 46736.829814386896)
    .check_interval(300.99967130971, 392.545588697268, 1249.0800632347068);

  regression_test::make(dvec{1, 1002, 2002, 3003},
                        dvec{59133, 59443, 59971, 60854})
    .with_lobf(0.5687614587449226, 58995.97028896512)
    .check_interval(203.72184550364616, 265.67233227980387, 845.369361314336);

  regression_test::make(dvec{1, 1002, 2002, 3003},
                        dvec{40847, 40858, 40954, 41195})
    .with_lobf(0.11393283502483618, 40792.3728817927)
    .check_interval(81.86463036395757, 106.75913143358696, 339.7075562216737);

  regression_test::make(dvec{1, 1001, 2002, 3002},
                        dvec{62854, 63093, 63286, 63420})
    .with_lobf(0.18902464248108664, 62879.42949931465)
    .check_interval(37.176711005083774, 48.480197695731974, 154.26398906781913);

  regression_test::make(dvec{1, 1002, 2002, 3002}, dvec{4483, 4503, 4590, 4638})
    .with_lobf(0.05518098313297275, 4470.631958580058)
    .check_interval(19.482792356187414, 25.40828088498815, 80.84914977603229);

  regression_test::make(dvec{1, 1002, 2002, 3003},
                        dvec{12803, 12804, 12806, 12808})
    .with_lobf(0.0016989605677344916, 12802.698161227263)
    .check_interval(0.38751798114251423, 0.5053596760620852,
                    1.6080544892295552);

  regression_test::make(dvec{1, 1002, 2002, 3003}, dvec{8080, 8081, 8083, 8085})
    .with_lobf(0.0016989605677344916, 8079.698161227263)
    .check_interval(0.3875179811432177, 0.5053596760630027, 1.6080544892324746);

  regression_test::make(dvec{1, 1002, 2003, 3003},
                        dvec{51330, 51982, 52913, 53189})
    .with_lobf(0.6503728203464637, 51376.47743063453)
    .check_interval(198.46777562944456, 258.82509417271524, 823.5814496575799);

  regression_test::make(dvec{1, 1002, 2002, 3003},
                        dvec{10853, 11400, 12379, 13156})
    .with_lobf(0.7883143075052707, 10762.951910127083)
    .check_interval(129.2727414210505, 168.5837403869376, 536.4334619112354);

  regression_test::make(dvec{2, 1002, 2003, 3003},
                        dvec{55662, 55901, 55924, 56153})
    .with_lobf(0.14953174469789274, 55685.32855359142)
    .check_interval(66.86488356225959, 87.21037471973891, 277.5034123582092);

  regression_test::make(dvec{1, 1002, 2003, 3003}, dvec{1661, 1731, 1885, 1969})
    .with_lobf(0.10772542515579804, 1649.6694800597024)
    .check_interval(24.83846403592538, 32.39224993989509, 103.07213930874617);

  regression_test::make(dvec{1, 1002, 2002, 3002}, dvec{6515, 6695, 6859, 6916})
    .with_lobf(0.13666423096178829, 6541.014491153134)
    .check_interval(45.76649721252338, 59.68590102683715, 189.9205370673958);

  regression_test::make(dvec{1, 1002, 2002, 3003},
                        dvec{12842, 12843, 12845, 12847})
    .with_lobf(0.0016989605677344916, 12841.698161227263)
    .check_interval(0.38751798114251423, 0.5053596760620852,
                    1.6080544892295552);

  regression_test::make(dvec{1, 1002, 2002, 3002}, dvec{8120, 8121, 8123, 8125})
    .with_lobf(0.001699430087033267, 8119.697880866798)
    .check_interval(0.38795653269498626, 0.5059494744730357,
                    1.6099312277731996);

  regression_test::make(dvec{1, 1001, 2002, 3002},
                        dvec{46868, 47040, 47439, 47450})
    .with_lobf(0.21442651590197523, 46877.288586373186)
    .check_interval(112.61690701018385, 146.8577980173928, 467.3015132913439);

  regression_test::make(dvec{1, 1002, 2002, 3003},
                        dvec{45784, 46021, 46275, 46978})
    .with_lobf(0.38337859233775773, 45688.66535430869)
    .check_interval(178.3074673940088, 232.52960726111013, 739.9092103048524);

  regression_test::make(dvec{1, 1002, 2002, 3002}, dvec{3584, 3769, 3826, 3837})
    .with_lobf(0.08158503674020037, 3631.479671075404)
    .check_interval(62.83882380245691, 81.95059806958534, 260.76680305742053);

  regression_test::make(dvec{1, 1002, 2002, 3003},
                        dvec{16032, 16277, 16408, 16622})
    .with_lobf(0.1899899360807287, 16049.385116006746)
    .check_interval(32.963780828204925, 42.98785194949918, 136.78734490330638);

  regression_test::make(dvec{1, 1001, 2002, 3002},
                        dvec{14355, 14495, 14643, 14779})
    .with_lobf(0.14194361671790234, 14354.871659498069)
    .check_interval(3.4231751275701088, 4.463983027143458, 14.204393992370482);

  regression_test::make(dvec{2, 1002, 2003, 3003},
                        dvec{12881, 12882, 12884, 12886})
    .with_lobf(0.001699340187981577, 12880.696741367557)
    .check_interval(0.3870792628771681, 0.5048588401462757, 1.6064608293454492);

  regression_test::make(dvec{1, 1002, 2002, 3003}, dvec{8159, 8160, 8162, 8164})
    .with_lobf(0.0016989605677344916, 8158.698161227263)
    .check_interval(0.3875179811432177, 0.5053596760630027, 1.6080544892324746);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{62578, 63251, 64076, 64672})
    .with_lobf(0.7104868513161826, 62577.27637103592)
    .check_interval(66.1066080007507, 86.21224482246498, 274.32736302508357);

  regression_test::make(dvec{1, 1002, 2002, 3002},
                        dvec{14068, 14119, 15103, 15371})
    .with_lobf(0.48912590109046056, 13930.7051780374)
    .check_interval(271.9063281148144, 354.6038079573781, 1128.349316920377);
}
